using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;
using GeometryHelper;

namespace PhingerPaint
{
    public class Game1 : Microsoft.Xna.Framework.Game
    {
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        Viewport viewport;
        List<ColorBlock> colorBlocks = new List<ColorBlock>();
        Color drawingColor = Color.Blue;

        Texture2D canvas;
        uint[] pixels;
        List<float> xCollection = new List<float>();

        bool ignoreTouchId;
        int touchIdToIgnore;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            // Frame rate is 30 fps by default for Windows Phone.
            TargetElapsedTime = TimeSpan.FromTicks(333333);

            // Set to portrait mode
            graphics.PreferredBackBufferWidth = 480;
            graphics.PreferredBackBufferHeight = 800;
        }

        protected override void Initialize()
        {
            Color[] colors = { Color.Red, Color.Green, Color.Blue, 
                               Color.Cyan, Color.Magenta, Color.Yellow,
                               Color.Black, new Color(0.2f, 0.2f, 0.2f), 
                                            new Color(0.4f, 0.4f, 0.4f), 
                                            new Color(0.6f, 0.6f, 0.6f), 
                                            new Color(0.8f, 0.8f, 0.8f), Color.White };

            foreach (Color clr in colors)
            {
                ColorBlock colorBlock = new ColorBlock(this)
                {
                    Color = clr,
                    IsSelected = drawingColor == clr

                };
                colorBlocks.Add(colorBlock);
                this.Components.Add(colorBlock);
            }

            viewport = this.GraphicsDevice.Viewport;
            int colorBlockSize = viewport.Width / (colorBlocks.Count / 2) - 2;
            int xColorBlock = 2;
            int yColorBlock = viewport.Height - 2 * colorBlockSize - 2;

            foreach (ColorBlock colorBlock in colorBlocks)
            {
                colorBlock.Destination = new Rectangle(xColorBlock, yColorBlock, colorBlockSize, colorBlockSize);
                xColorBlock += colorBlockSize + 2;

                if (xColorBlock + colorBlockSize > viewport.Width)
                {
                    xColorBlock = 2;
                    yColorBlock += colorBlockSize + 2;
                }
            }

            int canvasHeight = viewport.Height - 2 * colorBlockSize - 2;
            canvas = new Texture2D(this.GraphicsDevice, viewport.Width, canvasHeight);
            pixels = new uint[canvas.Width * canvas.Height];
            canvas.GetData<uint>(pixels);

            for (int y = 0; y < canvas.Height; y++)
                for (int x = 0; x < canvas.Width; x++)
                {
                    pixels[x + canvas.Width * y] = Color.Black.PackedValue;
                }

            canvas.SetData<uint>(pixels);

            base.Initialize();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
        }

        protected override void UnloadContent()
        {
        }

        protected override void Update(GameTime gameTime)
        {
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            TouchCollection touchLocations = TouchPanel.GetState();
            bool canvasNeedsUpdate = false;
            int yMin = 0, yMax = 0;

            foreach (TouchLocation touchLocation in touchLocations)
            {
                if (ignoreTouchId && touchLocation.Id == touchIdToIgnore)
                    continue;

                // Check for tap on ColorBlock
                if (touchLocation.State == TouchLocationState.Pressed)
                {
                    Vector2 position = touchLocation.Position;
                    ColorBlock newSelectedColorBlock = null;

                    foreach (ColorBlock colorBlock in colorBlocks)
                    {
                        Rectangle rect = colorBlock.Destination;

                        if (position.X >= rect.Left && position.X < rect.Right &&
                            position.Y >= rect.Top && position.Y < rect.Bottom)
                        {
                            drawingColor = colorBlock.Color;
                            newSelectedColorBlock = colorBlock;
                        }
                    }

                    if (newSelectedColorBlock != null)
                    {
                        foreach (ColorBlock colorBlock in colorBlocks)
                            colorBlock.IsSelected = colorBlock == newSelectedColorBlock;

                        ignoreTouchId = true;
                        touchIdToIgnore = touchLocation.Id;
                    }
                    else
                    {
                        ignoreTouchId = false;
                    }
                }

                // Check for drawing movement
                else if (touchLocation.State == TouchLocationState.Moved)
                {
                    TouchLocation prevTouchLocation;
                    touchLocation.TryGetPreviousLocation(out prevTouchLocation);

                    Vector2 point1 = prevTouchLocation.Position;
                    Vector2 point2 = touchLocation.Position;

                    // Sure hope touchLocation.Pressure comes back!
                    float radius = 12; //  48 * touchLocation.Pressure;

                    RoundCappedLine line = new RoundCappedLine(point1, point2, radius);

                    yMin = (int)(Math.Min(point1.Y, point2.Y) - radius);
                    yMax = (int)(Math.Max(point1.Y, point2.Y) + radius);

                    yMin = Math.Max(0, Math.Min(canvas.Height, yMin));
                    yMax = Math.Max(0, Math.Min(canvas.Height, yMax));

                    for (int y = yMin; y < yMax; y++)
                    {
                        xCollection.Clear();
                        line.GetAllX(y, xCollection);

                        if (xCollection.Count == 2)
                        {
                            int xMin = (int)(Math.Min(xCollection[0], xCollection[1]) + 0.5f);
                            int xMax = (int)(Math.Max(xCollection[0], xCollection[1]) + 0.5f);

                            xMin = Math.Max(0, Math.Min(canvas.Width, xMin));
                            xMax = Math.Max(0, Math.Min(canvas.Width, xMax));

                            for (int x = xMin; x < xMax; x++)
                            {
                                pixels[y * canvas.Width + x] = drawingColor.PackedValue;
                            }
                            canvasNeedsUpdate = true;
                        }
                    }
                }
            }

            if (canvasNeedsUpdate)
                canvas.SetData<uint>(pixels);

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            spriteBatch.Begin();
            spriteBatch.Draw(canvas, Vector2.Zero, Color.White);
            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
