using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Input.Touch;

namespace PhreeCell
{
    public partial class Game1 : Microsoft.Xna.Framework.Game
    {
        static readonly TimeSpan AutoMoveDuration = TimeSpan.FromSeconds(0.25);

        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;
        CongratulationsComponent congratsComponent;

        Texture2D cards;
        Texture2D surface;
        Rectangle[] cardSpots = new Rectangle[16];

        public Matrix DisplayMatrix { set; get; }
        Matrix inverseMatrix;

        CardInfo[] deck = new CardInfo[52];
        List<CardInfo>[] piles = new List<CardInfo>[8];
        CardInfo[] holds = new CardInfo[4];
        List<CardInfo>[] finals = new List<CardInfo>[4];

        CardInfo touchedCard;
        Vector2 touchedCardPosition;
        object touchedCardOrigin;
        int touchedCardOriginIndex;
        int touchedCardTouchId;

        public Game1()
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            // Frame rate is 30 fps by default for Windows Phone.
            TargetElapsedTime = TimeSpan.FromTicks(333333);

            congratsComponent = new CongratulationsComponent(this);
            congratsComponent.Enabled = false;
            this.Components.Add(congratsComponent);
        }

        protected override void Initialize()
        {
            // Initialize deck
            for (int suit = 0; suit < 4; suit++)
                for (int rank = 0; rank < 13; rank++)
                {
                    CardInfo cardInfo = new CardInfo(suit, rank);
                    deck[suit * 13 + rank] = cardInfo;
                }

            // Create the List objects for the 8 piles
            for (int pile = 0; pile < 8; pile++)
                piles[pile] = new List<CardInfo>();

            // Create the List objects for the 4 finals
            for (int final = 0; final < 4; final++)
                finals[final] = new List<CardInfo>();

            base.Initialize();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // Load large bitmap containing cards
            cards = this.Content.Load<Texture2D>("cards");

            // Create the 16 rectangular areas for the cards and the bitmap surface
            CreateCardSpots(cardSpots);
            surface = CreateSurface(this.GraphicsDevice, cardSpots);

            // Start the game!
            Replay();
        }

        protected override void UnloadContent()
        {
        }

        protected override void Update(GameTime gameTime)
        {
            if (GamePad.GetState(PlayerIndex.One).Buttons.Back == ButtonState.Pressed)
                this.Exit();

            // Process auto-move card and perhaps initiate next auto-move
            bool checkForNextAutoMove = false;

            foreach (List<CardInfo> final in finals)
                foreach (CardInfo cardInfo in final)
                {
                    if (cardInfo.AutoMoveTime > TimeSpan.Zero)
                    {
                        cardInfo.AutoMoveTime -= gameTime.ElapsedGameTime;

                        if (cardInfo.AutoMoveTime <= TimeSpan.Zero)
                        {
                            cardInfo.AutoMoveTime = TimeSpan.Zero;
                            checkForNextAutoMove = true;
                        }
                        cardInfo.AutoMoveInterpolation = (float)cardInfo.AutoMoveTime.Ticks / AutoMoveDuration.Ticks;
                    }
                }

            if (checkForNextAutoMove && !AnalyzeForAutoMove() && HasWon())
            {
                congratsComponent.Enabled = true;
            }

            TouchCollection touchLocations = TouchPanel.GetState();

            // Try to pick up a card
            if (touchedCard == null)
            {
                foreach (TouchLocation touchLocation in touchLocations)
                {
                    // Finger pressed to screen
                    if (touchLocation.State == TouchLocationState.Pressed)
                    {
                        Vector2 position = Vector2.Transform(touchLocation.Position, inverseMatrix);

                        // Replay pressed?
                        if ((position - centerReplay).Length() < radiusReplay)
                        {
                            congratsComponent.Enabled = false;
                            Replay();
                        }

                        // Now try to pick up a card
                        else if (TryPickUpCard(position))
                        {
                            touchedCardTouchId = touchLocation.Id;
                        }
                    }
                }
            }
            else
            {
                foreach (TouchLocation touchLocation in touchLocations)
                {
                    // If a card is being moved, only process IDs for that card
                    if (touchLocation.Id == touchedCardTouchId)
                    {
                        if (touchLocation.State == TouchLocationState.Moved)
                        {
                            TouchLocation previousTouchLocation;
                            touchLocation.TryGetPreviousLocation(out previousTouchLocation);
                            Vector2 moveVector = touchLocation.Position - previousTouchLocation.Position;

                            Vector2 position = Vector2.Transform(touchedCardPosition, DisplayMatrix);
                            position += moveVector;
                            touchedCardPosition = Vector2.Transform(position, inverseMatrix);
                        }
                        // The finger has been lifted from the screen: Try to set down the card
                        else
                        {
                            if (TryPutDownCard(touchedCard))
                            {
                                CalculateDisplayMatrix();

                                // This will not happen!
                                if (!AnalyzeForAutoMove() && HasWon())
                                {
                                    congratsComponent.Enabled = true;
                                }
                            }

                            touchedCard = null;
                        }
                    }
                }
            }

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            spriteBatch.Begin(SpriteSortMode.Immediate, null, null, null, null, null, DisplayMatrix);
            spriteBatch.Draw(surface, Vector2.Zero, Color.White);

            // Draw holds
            for (int hold = 0; hold < 4; hold++)
            {
                CardInfo cardInfo = holds[hold];

                if (cardInfo != null)
                {
                    Rectangle source = GetCardTextureSource(cardInfo);
                    Vector2 destination = new Vector2(cardSpots[hold].X, cardSpots[hold].Y);
                    spriteBatch.Draw(cards, destination, source, Color.White);
                }
            }

            // Draw piles
            for (int pile = 0; pile < 8; pile++)
            {
                Rectangle cardSpot = cardSpots[pile + 8];

                for (int card = 0; card < piles[pile].Count; card++)
                {
                    CardInfo cardInfo = piles[pile][card];
                    Rectangle source = GetCardTextureSource(cardInfo);
                    Vector2 destination = new Vector2(cardSpot.X, cardSpot.Y + card * yOverlay);
                    spriteBatch.Draw(cards, destination, source, Color.White);
                }
            }

            // Draw finals including all previous cards (for auto-move)
            for (int pass = 0; pass < 2; pass++)
            {
                for (int final = 0; final < 4; final++)
                {
                    for (int card = 0; card < finals[final].Count; card++)
                    {
                        CardInfo cardInfo = finals[final][card];

                        if (pass == 0 && cardInfo.AutoMoveInterpolation == 0 ||
                            pass == 1 && cardInfo.AutoMoveInterpolation != 0)
                        {
                            Rectangle source = GetCardTextureSource(cardInfo);
                            Vector2 destination = new Vector2(cardSpots[final + 4].X,
                                                              cardSpots[final + 4].Y) +
                                         cardInfo.AutoMoveInterpolation * cardInfo.AutoMoveOffset;
                            spriteBatch.Draw(cards, destination, source, Color.White);
                        }
                    }
                }
            }

            // Draw touched card
            if (touchedCard != null)
            {
                Rectangle source = GetCardTextureSource(touchedCard);
                spriteBatch.Draw(cards, touchedCardPosition, source, Color.White);
            }

            spriteBatch.End();

            base.Draw(gameTime);
        }
    }
}
